# -*- coding: utf-8 -*-

import os.path
import collections
import importlib

from sett import optional_import, defaults, ROOT
from sett.utils.dispatch import Dispatcher
from sett.utils.watcher import Watcher as BaseWatcher
from paver.easy import task, info, debug, consume_args, path, error
from paver.deps.six import string_types


observers = optional_import('watchdog.observers')
events = optional_import('watchdog.events')
libsass = optional_import('sass', 'libsass')


# Time to wait for additional events in seconds
DEBOUNCE_WAIT = 0.250


class Watcher(BaseWatcher):
    def __init__(self, builder):
        assert isinstance(builder, Sass)
        super(Watcher, self).__init__(builder, builder._paths, {'.css', '.scss'})


class SassDispatcher(Dispatcher):
    def __init__(self):
        self._sass = Sass.default()

    @Dispatcher.auto
    @Dispatcher.on('watch', -1)
    def compile(self):
        """Launches the compilation of all sass files"""
        debug('Building all files with %s', self._sass)
        self._sass()

    def watch(self):
        """Watches the sass source directories and launches compilation as needed"""
        debug('Building/watching with %s', self._sass)
        try:
            watcher = BaseWatcher(self._sass, self._sass._paths, {'.css', '.scss'})
            info('Starting watch %s', watcher)
            watcher()
        except KeyboardInterrupt:
            pass


@task
@consume_args
def sass(args):
    """
    Usage: sass [compile|watch]

    Shortcut to sassc_watch and sassc_compile.
    """
    sd = SassDispatcher()
    sd(*args)


@task
def sassc_compile():
    """Compile the sass files"""
    sd = SassDispatcher()
    sd.compile()


@task
def sassc_watch():
    """Watch the filesystem for modifications and run sass if needed"""
    sd = SassDispatcher()
    sd.watch()


class DependencyTracker(object):
    def __init__(self):
        self._deps = set()

    def __contains__(self, file):
        return file in self._deps

    def __call__(self, import_string):
        self._deps.add(import_string)
        return [(import_string, )]


class BaseSass(object):
    @classmethod
    def get_default_paths(self):
        sp = defaults.SASS_PATH
        if not sp:
            return []
        if isinstance(sp, string_types):
            return sp.split(':')
        if isinstance(sp, collections.Iterable):
            return [ROOT.joinpath(d) for d in sp]

        bc = ROOT.joinpath('bower_components/'),
        if bc.isdir():
            return [bc]
        return []

    @classmethod
    def get_default_functions(self):
        functions = []
        functions_names = ([defaults.SASS_FUNCTIONS]
                           if isinstance(defaults.SASS_FUNCTIONS, string_types) else
                           defaults.SASS_FUNCTIONS)

        for sassf in functions_names:
            if isinstance(sassf, string_types):
                module = importlib.import_module(sassf)
                functions.extend(f for f in vars(module).values() if isinstance(f, libsass.SassFunction))
            elif isinstance(sassf, libsass.SassFunction):
                functions.append(sassf)
            else:
                raise TypeError('unexpected %s in defaults.SASS_FUNCTIONS' % sassf)

        return functions

    @classmethod
    def default(cls):
        return cls(
            ROOT.joinpath(defaults.SASS_SRC_DIR, 'scss'),
            ROOT.joinpath(defaults.SASS_BUILD_DIR),
            cls.get_default_paths(),
            cls.get_default_functions(),
        )

    def __init__(self, src, dest, include_paths, functions):
        self._src = src
        self._dest = dest
        self._deps = {}
        self._functions = functions

        paths = [self._src]
        paths.extend(include_paths)
        # Ensure trailing slashes so it does not mess with deps computation
        self._paths = [os.path.join(p, '') for p in paths]

    def __repr__(self):
        return 'Sass({} -> {}, {})'.format(
            self._src,
            self._dest,
            ':'.join(self._paths),
        )

    def get_compile_kwargs(self):
        return {
            'importers': [],
            'output_style': defaults.SASS_OUTPUT_STYLE,
            'custom_functions': self._functions,
            'include_paths': self._paths,
        }

    def __call__(self, filename=None):
        if filename is None:
            self._build_all()
            return
        if isinstance(filename, (list, set, tuple, frozenset)):
            for fname in filename:
                self(fname)
            return

        filename = path(filename)
        if filename.basename().startswith('_'):
            if not self._deps:
                self._build_all()
                return

            dep_names = {
                os.path.join(filename.dirname()[len(p):], filename.basename().stripext().lstrip('_'))
                for p in self._paths
                if filename.startswith(p)
            }

            if not dep_names:
                return

            for main_file, deps in self._deps.items():
                if any(dep_name in deps for dep_name in dep_names):
                    self._build_file(main_file)
            return

        self._build_file(filename)

    def _build_file(self, filename):
        assert isinstance(filename, string_types), '%s is not a string type' % filename

        kwargs = self.get_compile_kwargs()
        dep_tracker = DependencyTracker()
        kwargs['importers'].append((0, dep_tracker))

        infile = self._src.joinpath(filename)
        try:
            result = libsass.compile(filename=infile, **kwargs)
        except Exception as e:
            error('Cannot build %s: %s', infile, e)
            raise

        self._deps[infile] = dep_tracker
        relative_infile = os.path.relpath(infile, self._src)

        outfile = self._dest.joinpath(relative_infile).stripext() + '.css'
        if not outfile.parent.isdir():
            outfile.parent.makedirs_p()

        info('Build %s -> %s', filename, outfile)
        with open(outfile, 'wb') as out_stream:
            out_stream.write(result.encode('utf-8'))

    def _build_all(self):
        for filename in path(self._src).walkfiles():

            if filename.ext in {'.scss', '.sass'} and not filename.basename().startswith('_'):
                self._build_file(filename)


class Sass(BaseSass):
    def get_compile_kwargs(self):
        kw = super(Sass, self).get_compile_kwargs()
        kw.setdefault('importers', []).append((1, Sass.css_import))
        return kw

    @staticmethod
    def css_import(path):
        if not path.startswith('CSS:'):
            return None

        s = path[4:] + '.css'
        return [(s, '',)]
