# -*- coding: utf-8 -*-

import time
import os.path
import threading

from paver.deps.six import moves
from paver.easy import debug

from sett import optional_import


events = optional_import('watchdog.events')
observers = optional_import('watchdog.observers')


# Time to wait for additional events in seconds
DEBOUNCE_WAIT = 0.250


class Watcher(object):
    """Watchdog compatible event dispatcher. It can be called in the same
    thread or started in another thread.

    >>> with Watcher(Sass.default()):
    ...     call_task('runserver')  # Running an http server and building at the same time

    >>> w = Watcher(Sass.default())
    >>> w()  # Blocks until it's stopped
    """
    def __init__(self, builder, paths, extensions, debounce_wait=DEBOUNCE_WAIT):
        self.observer = observers.Observer()
        self.queue = moves.queue.Queue()
        self._thread = None
        self._stop = object()

        self.watches = {}
        for filepath in paths:
            watch = self.observer.schedule(EventDispatcher(self.queue, extensions),
                                           filepath,
                                           recursive=True)
            debug('Watching %s', filepath)
            self.watches[filepath] = watch
        self.builder = builder
        self._debounce_wait = debounce_wait

    def start(self):
        self.start_observer()
        assert self._thread is None
        self._thread = threading.Thread(target=self.run)
        self._thread.start()

    def stop(self):
        self.stop_observer()
        self.queue.put(self._stop)
        self._thread.join()

    def __call__(self):
        self.start_observer()
        debug('Processing loop')
        try:
            self.run()
        finally:
            self.stop_observer()

    def run(self):
        while True:
            path = self.queue.get()
            if path is self._stop:
                break

            paths = {path}
            first_event = time.time()
            try:
                # Debounce: capture all events occuring 250ms after the fist
                delay = self._debounce_wait
                while delay > 0:
                    paths.add(self.queue.get(timeout=delay))
                    delay = first_event - time.time() + self._debounce_wait

            except moves.queue.Empty:
                pass

            if self._stop in paths:
                break

            self.builder(paths)

    def start_observer(self):
        debug('Start observer')
        self.observer.start()

    def stop_observer(self):
        debug('Stop observer')
        self.observer.stop()
        self.observer.join()

    def __enter__(self):
        self.start()

    def __exit__(self, exc_value, exc_type, tb):
        self.stop()


class EventDispatcher(object):
    """Watchdog compatible event dispatcher. It calls the builder when a file
    is modified and have the right extension"""
    def __init__(self, queue, extensions):
        self.extensions = extensions
        self.queue = queue

    def dispatch(self, event):
        #  events.FileCreatedEvent triggers FileModifiedEvent
        debug(event)
        if isinstance(event, (events.FileModifiedEvent)) and event.src_path in self:
            self.queue.put(event.src_path)

    def __contains__(self, path):
        stem, ext = os.path.splitext(path)
        if ext not in self.extensions:
            return False
        return True
